@@ -1,7 +1,8 @@
 /**********************************************************************
  *  cgi.c -- libcgi
  *
- * Copyright 1994, 1996 by the Massachusetts Institute of Technology
+ * Copyright 1994, 1996, 1998
+ * by the Massachusetts Institute of Technology
  * For copying and distribution information, please see the file
  * <mit-copyright.h>.
  **********************************************************************/
@@ -17,10 +18,16 @@
 #include <time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 #include <fcntl.h>
 #include "cgi.h"
+#ifdef ENABLE_OWNER_BOUNCE
+#include <pwd.h>
+#endif /* ENABLE_OWNER_BOUNCE */
 
 char *cgi_query=(char*)0;
+time_t received_time;
+char datebuf[80];
 
 int cgi_alloc_form(formp)
      cgi_form *formp;
@@ -133,7 +140,35 @@ cgi_free_form(formp)
   return;
 }
 
-/* converts cr/lf pairs to lf, which Unix will handle better */
+/* Get the full pathname of a file parallel to the PATH_TRANSLATED file */
+/* Return 0 on success, nonzero on failure. */
+
+int
+cgi_parallel_fname(basename, buf, buflen)
+     char *basename, *buf;
+     int buflen;
+{
+  int dlen;
+  char *env_var, *dir_end;
+
+  /* Get directory */
+  env_var = getenv("PATH_TRANSLATED");
+  if (!env_var)
+    return 1;
+  dir_end = strrchr(env_var, '/');
+  if (!dir_end)
+    return 1;
+  dlen = dir_end - env_var + 1;
+
+  if (1 + dlen + strlen(basename) > buflen)
+    return 1;
+
+  strncpy(buf, env_var, dlen);
+  strcpy(buf+dlen, basename);
+  return 0;
+}
+
+/* converts cr/lf pairs or cr by itself to lf, which Unix will handle better */
 void
 cgi_fix_crlf(string)
      char *string;
@@ -178,6 +213,13 @@ cgi_parse_form(formp)
   char hex_string[3];
   int parse_state=0;
   int hex_int, i;
+  struct tm *timeptr;
+
+  /* Get date-time (RFC 822 sec. 5, RFC 1123 sec. 5.2.14) */
+  /* FIXME: not compliant if lang is not English or TZ unknown */
+  time(&received_time);
+  timeptr=localtime(&received_time);
+  strftime(datebuf, 79, "%a, %d %b %Y %H:%M:%S %Z", timeptr);
 
   /* initialize variables */
   field_ptr=formp->fields;
@@ -201,19 +243,6 @@ cgi_parse_form(formp)
 	  /* we're at the 'v' in "name=value&name2=value2..." */
 	  field_ptr->value = to_ptr;
 	  parse_state++;
-
-	  if (from_ptr[0] == '&')
-	    {
-	      /* we're at the '&' in "name=&..." - value is "" */
-	      if (cgi_required(field_ptr->name))
-		{
-		  formp->errcond=1;
-		  sprintf(formp->errmsg,
-			  "400 Required field left blank: %-.512s",
-			  field_ptr->name);
-		  return(1);
-		}
-	    }
 	}
       if (from_ptr[0] == '%' && cgi_ishex((int)from_ptr[1]))
 	{
@@ -259,25 +288,27 @@ cgi_parse_form(formp)
   *to_ptr='\0';
   if (parse_state==2) field_ptr->value = to_ptr;
 
-  /* fix CR/LF; check required fields */
+  /* fix CR/LF */
   for (i=0; i<formp->nfields; i++)
-    {
-      cgi_fix_crlf(formp->fields[i].value);
+    cgi_fix_crlf(formp->fields[i].value);
 
-      if (cgi_required(formp->fields[i].name)
-	  && !cgi_nonblank(formp->fields[i].value))
-	{
-	  formp->errcond=1;
-	  sprintf(formp->errmsg,
-		  "400 Required field left blank: %-.512s",
-		  formp->fields[i].name);
-	  return(1);
-	}
+  /* Add "special" output variables */
+  if (formp->maxfields - formp->nfields > 4)
+    {
+      formp->fields[formp->nfields].name = CGI_ERRMSG;
+      formp->fields[formp->nfields++].value = formp->errmsg;
+      formp->fields[formp->nfields].name = CGI_ERRINFO;
+      formp->fields[formp->nfields++].value = formp->errinfo;
+      formp->fields[formp->nfields].name = CGI_RELEASE;
+      formp->fields[formp->nfields++].value = CGIEMAIL_RELEASE;
+      formp->fields[formp->nfields].name = CGI_DATE;
+      formp->fields[formp->nfields++].value = datebuf;
     }
 
   return(0);
 }
 
+/* only for debugging purposes */
 void
 cgi_print_form(formp, outstream)
      cgi_form *formp;
@@ -308,6 +339,77 @@ cgi_value(formp, name)
   return("");
 }
 
+/* Function to translate an integer character into an HTML entity string */
+char *
+cgi_char2entity(c)
+     int c;
+{
+  static char retval[2];
+
+  switch (c)
+    {
+    case (int)'&':
+      return("&amp;");
+    case (int)'<':
+      return("&lt;");
+    case (int)'>':
+      return("&gt;");
+    case (int)'"':
+      return("&quot;");
+    default:
+      sprintf(retval, "%c", c);
+      return(retval);
+    }
+}
+
+/*
+ * Non-alphanumeric chars that don't need %nn encoding in form values.
+ * See also: http://www.w3.org/Addressing/URL/5_BNF.html
+ */
+static char url_safe[]="/:$-_@.";
+
+void
+cgi_string2url(instring, outstream)
+     char *instring;
+     FILE *outstream;
+{
+  while (*instring)
+    {
+      if (*instring == ' ')
+	fputc('+', outstream);
+      else
+	{
+	  if (isalnum(*instring) || strchr(url_safe, *instring))
+	    fputc(*instring, outstream);
+	  else
+	    fprintf(outstream, "%%%02x", (int)*instring);
+	}
+      instring++;
+    }
+  return;
+}
+
+void
+cgi_string2html(instring, outstream)
+     char *instring;
+     FILE *outstream;
+{
+  while (*instring)
+    fputs(cgi_char2entity(*instring++), outstream);
+  return;
+}
+  
+void
+cgi_stream2html(instream, outstream)
+     FILE *instream, *outstream;
+{
+  int c;
+
+  while ((c=fgetc(instream)) != EOF)
+    fputs(cgi_char2entity(c), outstream);
+  return;
+}
+
 /* Output the value(s) of a CGI field name; return number found */
 int
 cgi_output_value(formp, name, formatstr, outstream)
@@ -318,16 +420,26 @@ cgi_output_value(formp, name, formatstr, outstream)
 {
   int i, nfound=0;
 
-  for (i=0; i<formp->nfields; i++)
+  for (i=0; i < formp->nfields; i++)
     {
       if (!strcmp(formp->fields[i].name, name))
 	{
 	  if (nfound) fputc(' ', outstream);
 	  if (formatstr)
-	    fprintf(outstream, formatstr, formp->fields[i].value);
+	    {
+	      if (!strcmp(formatstr, CGI_HTMLENCODE))
+		cgi_string2html(formp->fields[i].value, outstream);
+	      else
+		{
+		  if (!strcmp(formatstr, CGI_URLENCODE))
+		    cgi_string2url(formp->fields[i].value, outstream);
+		  else
+		    fprintf(outstream, formatstr, formp->fields[i].value);
+		}
+	    }
 	  else
 	    fputs(formp->fields[i].value, outstream);
-	  nfound++;
+	  if (cgi_nonblank(formp->fields[i].value)) nfound++;
 	}	  
     }
   return(nfound);
@@ -369,8 +481,9 @@ cgi_template_fill(formp, templatefile)
   if (!tfp)
     {
       formp->errcond=1;
-      sprintf(formp->errmsg, "500 Could not open %-.512s", templatefile);
+      strcpy(formp->errmsg, "500 Could not open template");
       cgi_concat_errno(formp->errmsg);
+      strncpy(formp->errinfo, templatefile, CGI_ERRMSG_MAX);
       return(1);
     }
 
@@ -380,20 +493,37 @@ cgi_template_fill(formp, templatefile)
   if (!formp->tmpf)
     {
       formp->errcond=1;
-      strcpy(formp->errmsg, "503 Could not open temporary file.");
-      cgi_concat_errno(formp->errmsg);
+      strcpy(formp->errmsg, "500 Could not open temporary file.");
+      /* Try to make the error message more informative */
+      if (access(P_tmpdir, W_OK))
+	{
+	  strcpy(formp->errinfo, P_tmpdir);
+	  cgi_concat_errno(formp->errinfo);
+	}
+      return(1);
+    }
+
+  while ((inchar=fgetc(tfp)) != EOF
+	 && isspace(inchar))
+    { /* Skip leading whitespace */ }
+
+  if (inchar == EOF)
+    {
+      formp->errcond=1;
+      strcpy(formp->errmsg, "500 Empty template file");
+      strncpy(formp->errinfo, templatefile, CGI_ERRMSG_MAX);
       return(1);
     }
 
   /*
    * parsing states:
-   * 0: echo
-   * 1: quote
-   * 2: variable
-   * 3: format
+   * 0: echo       sending literally except for CRLF conversion or \, [
+   * 1: quote      sending literally no matter what
+   * 2: variable   reading variable name
+   * 3: format     reading format
    */
-  
-  while ((inchar=fgetc(tfp)) != EOF)
+
+  do  
     {
       if (parse_state==1)	/* quote */
 	{
@@ -403,6 +533,13 @@ cgi_template_fill(formp, templatefile)
 	}
       if (parse_state==0)	/* echo */
 	{
+	  /* Convert CR or CRLF to LF */
+	  if (inchar == (int)'\r')
+	    {
+	      fputc('\n', formp->tmpf);
+	      inchar=fgetc(tfp);
+	      if (inchar == (int)'\n') continue;
+	    }
 	  if (inchar == (int)'\\')
 	    {
 	      parse_state=1;
@@ -468,9 +605,9 @@ cgi_template_fill(formp, templatefile)
 	          if (!nfound && cgi_required(varname))
 		    {
 		      formp->errcond=1;
-		      sprintf(formp->errmsg,
-			      "400 Required field left blank: %-.512s",
-			      varname);
+		      strcpy(formp->errmsg,
+			      "400 Required field left blank");
+		      strcpy(formp->errinfo, varname);
 		      return(1);
 		    }
 		  substitutions += nfound;
@@ -483,8 +620,9 @@ cgi_template_fill(formp, templatefile)
 	    {
 	      fclose(tfp);
 	      formp->errcond=1;
-	      sprintf(formp->errmsg, "403 Variable Name too long (%-.32s...)",
-		      varname);
+	      strcpy(formp->errmsg, "403 Variable Name too long");
+	      strncpy(formp->errinfo, varname, varnamelen);
+	      strcpy(formp->errinfo + varnamelen, "...");
 	      return(1);
 	    }
 	  continue;
@@ -502,13 +640,15 @@ cgi_template_fill(formp, templatefile)
 	    {
 	      fclose(tfp);
 	      formp->errcond=1;
-	      sprintf(formp->errmsg, "403 Format String too long (%-.32s...)",
-		      formatstr);
+	      strcpy(formp->errmsg, "403 Format String too long");
+	      strncpy(formp->errinfo, formatstr, formatlen);
+	      strcpy(formp->errinfo + formatlen, "...");
 	      return(1);
 	    }
 	  continue;
 	}
-    }
+    } while ((inchar=fgetc(tfp)) != EOF);
+
   fclose(tfp);
 
   /* If there were no variable substitutions in this template,
@@ -516,8 +656,8 @@ cgi_template_fill(formp, templatefile)
   if (!substitutions)
     {
       formp->errcond=1;
-      sprintf(formp->errmsg, "403 No variable substitutions in \"%-.512s\"",
-		      templatefile);
+      strcpy(formp->errmsg, "403 No variable substitutions in template");
+      strncpy(formp->errinfo, templatefile, CGI_ERRMSG_MAX);
       return(1);
     }
 
@@ -562,10 +702,21 @@ cgi_mail_template(formp, templatefile)
   int retval;
   char buf[BUFSIZ], command[BUFSIZ];
   int nbytes;
-  char *envvar, datebuf[80], errorfile[L_tmpnam];
+  char *envvar, errorfile[L_tmpnam];
   int old_stdout, old_stderr, errorfd;
-  time_t rightnow;
-  struct tm *timeptr;
+#ifdef ENABLE_OWNER_BOUNCE
+  char *owner = NULL;
+  struct passwd *pw;
+  struct stat st;
+
+  /* Get owner of template */
+  if (0 == stat(templatefile, &st))
+    {
+      pw = getpwuid(st.st_uid);
+      if (pw)
+	owner = pw->pw_name;
+    }
+#endif /* ENABLE_OWNER_BOUNCE */
 
   /* fill in template */
   retval=cgi_template_fill(formp, templatefile);
@@ -587,6 +738,10 @@ cgi_mail_template(formp, templatefile)
   strcat(strcpy(command, PATH_SENDMAIL), " -oi -t");
   if (strstr(cgi_value(formp, "cgiemail-mailopt"), "sync"))
     strcat(command, " -odi");	/* synchronous delivery */
+#ifdef ENABLE_OWNER_BOUNCE
+  if (owner)
+    strcat(strcat(command, " -f "), owner); /* envelope set to owner */
+#endif /* ENABLE_OWNER_BOUNCE */
   mypipe = popen(command, "w");
 
   /* Restore stdout/stderr */
@@ -603,8 +758,9 @@ cgi_mail_template(formp, templatefile)
   if (!mypipe)
     {
       formp->errcond=1;
-      strcpy(formp->errmsg, "500 Could not open sendmail pipe.");
+      strcpy(formp->errmsg, "500 Could not open sendmail pipe");
       cgi_concat_errno(formp->errmsg);
+      strncpy(formp->errinfo, command, CGI_ERRMSG_MAX);
       return(1);
     }
 
@@ -623,14 +779,9 @@ cgi_mail_template(formp, templatefile)
 	  fputs(" by ", mypipe);
 	  fputs(envvar, mypipe);
 	}
-      fputs(" with HTTP;\n", mypipe);
-
-      /* Put date-time (RFC 822 sec. 5, RFC 1123 sec. 5.2.14) */
-      /* FIXME: not compliant if lang is not English or TZ unknown */
-      time(&rightnow);
-      timeptr=localtime(&rightnow);
-      strftime(datebuf, 79, "\t%a, %d %b %Y %H:%M:%S %Z\n", timeptr);
+      fputs(" with HTTP;\n\t", mypipe);
       fputs(datebuf, mypipe);
+      fputs("\n", mypipe);
     }
 
 #ifdef ENABLE_XHEADERS
@@ -659,7 +810,14 @@ cgi_mail_template(formp, templatefile)
   } while (nbytes == BUFSIZ);
 
   /* close */
-  retval=pclose(mypipe);
+  retval = pclose(mypipe);
+
+  /*
+   * Get exit status.  Note that on SunOS 4, WEXITSTATUS(stat)
+   * only works if &stat is legal, so WEXITSTATUS(pclose(mypipe))
+   * won't compile.
+   */
+  retval = WEXITSTATUS(retval);
 
   if (retval)
     {
@@ -699,6 +857,43 @@ cgi_output_failure(formp, msg)
      cgi_form *formp;
      char *msg;
 {
+  char *failure;
+
+  /* maybe use customized failure message */
+  failure=cgi_value(formp, CGI_FAILURE);
+  if (failure && *failure)
+    {
+      if (strpbrk(failure, ":/"))
+	{			/* Probably a URL.  Redirect. */
+	  cgi_redirect(failure);
+	  return;
+	}
+      else
+	{			/* Probably a template file.  Use it. */
+	  char buf[BUFSIZ];
+	  int nbytes;
+
+	  if (0 == cgi_parallel_fname(failure, buf, BUFSIZ)
+	      && 0 == cgi_template_fill(formp, buf))
+	    {
+	      rewind(formp->tmpf);
+	      puts("Content-Type: text/html\r\n\r");
+	      do
+		{
+		  nbytes = fread(buf, sizeof(char), BUFSIZ, formp->tmpf);
+		  if (nbytes)
+		    {
+		      if (!fwrite(buf, sizeof(char), nbytes, stdout))
+			break;
+		    }
+		}
+	      while (nbytes == BUFSIZ);
+	      return;
+	    }
+	}
+    }
+
+  /* use generic failure message */
   puts("Content-Type: text/html\r");
   printf("Status: %s\r\n\r\n", formp->errmsg);
   puts("<HEAD><TITLE>Error</TITLE></HEAD>");
@@ -720,37 +915,12 @@ cgi_output_failure(formp, msg)
   return;
 }
 
-/* Function to translate an integer character into an HTML entity string */
-char *
-cgi_char2entity(c)
-     int c;
-{
-  static char retval[2];
-
-  switch (c)
-    {
-    case (int)'&':
-      return("&amp;");
-    case (int)'<':
-      return("&lt;");
-    case (int)'>':
-      return("&gt;");
-    default:
-      sprintf(retval, "%c", c);
-      return(retval);
-    }
-}
-  
-void
-cgi_ascii2html(instream, outstream)
-     FILE *instream, *outstream;
-{
-  int c;
-
-  while ((c=fgetc(instream)) != EOF)
-    fputs(cgi_char2entity(c), outstream);
-  return;
-}
+/*
+ * See RFC 2068, section 10.3 for information on how HTTP redirection
+ * should work.  As of 1997, an explicit Status: 303 message doesn't
+ * do the job.  For now, just hope httpd does the right thing with
+ * Location (see <URL:http://hoohoo.ncsa.uiuc.edu/cgi/out.html>).
+ */
 
 void
 cgi_redirect(url)
@@ -771,8 +941,38 @@ cgi_output_success(formp, msg)
   success=cgi_value(formp, CGI_SUCCESS);
   if (success && *success)
     {
-      cgi_redirect(success);
-      return;
+      if (strpbrk(success, ":/"))
+	{			/* Probably a URL.  Redirect. */
+	  cgi_redirect(success);
+	  return;
+	}
+      else
+	{			/* Probably a template file.  Use it. */
+	  char buf[BUFSIZ];
+	  int nbytes;
+
+	  if (0 == cgi_parallel_fname(success, buf, BUFSIZ)
+	      && 0 == cgi_template_fill(formp, buf))
+	    {
+	      rewind(formp->tmpf);
+	      puts("Content-Type: text/html\r\n\r");
+	      do
+		{
+		  nbytes = fread(buf, sizeof(char), BUFSIZ, formp->tmpf);
+		  if (nbytes)
+		    {
+		      if (!fwrite(buf, sizeof(char), nbytes, stdout))
+			break;
+		    }
+		}
+	      while (nbytes == BUFSIZ);
+	      return;
+	    }
+	  else
+	    {
+	      return cgi_output_failure(formp, msg);
+	    }
+	}
     }
 
   /* use generic success message */
@@ -781,7 +981,7 @@ cgi_output_success(formp, msg)
   printf("<BODY>%s<P><HR>", msg);
   puts("<PRE>");
   rewind(formp->tmpf);
-  cgi_ascii2html(formp->tmpf, stdout);
+  cgi_stream2html(formp->tmpf, stdout);
   if (*cgi_value(formp, "debug-source"))
     {
       puts("<HR>");
@@ -806,8 +1006,7 @@ cgi_standard_email()
   template_filename = getenv("PATH_TRANSLATED");
   if (!template_filename || !(*template_filename))
     {
-      /* FIXME: shouldn't hard-code this URL */
-      cgi_redirect("http://web.mit.edu/wwwdev/cgiemail/nopath.html");
+      cgi_redirect(CGI_NOPATH);
       return(1);
     }
 
@@ -834,8 +1033,7 @@ cgi_standard_echo()
   template_filename = getenv("PATH_TRANSLATED");
   if (!template_filename || !(*template_filename))
     {
-      /* FIXME: shouldn't hard-code this URL */
-      cgi_redirect("http://web.mit.edu/wwwdev/cgiemail/nopath.html");
+      cgi_redirect(CGI_NOPATH);
       return(1);
     }
 
@@ -852,3 +1050,78 @@ cgi_standard_echo()
   cgi_free_form(&form);
   return(0);
 }
+
+int
+cgi_standard_file()
+{
+  cgi_form form;
+  char *template_filename, incoming_filename[BUFSIZ];
+  int retval;
+
+  template_filename = getenv("PATH_TRANSLATED");
+  if (!template_filename || !(*template_filename))
+    {
+      cgi_redirect(CGI_NOPATH);
+      return(1);
+    }
+
+  if (cgi_alloc_form(&form) ||
+      cgi_parse_form(&form) ||
+      cgi_template_fill(&form, template_filename))
+    {
+      cgi_output_failure(&form, "Form was not processed due to an error.");
+      cgi_free_form(&form);
+      return(1);
+    }
+
+  if (0 == cgi_parallel_fname(CGI_INFILE, incoming_filename, BUFSIZ))
+    {
+      if (!access(incoming_filename, W_OK))
+	{
+	  FILE *fp;
+	  char buf[BUFSIZ];
+	  int nbytes;
+
+	  rewind(form.tmpf);
+	  fp = fopen(incoming_filename, "a");
+	  if (fp)
+	    {
+	      do
+		{
+		  if ((nbytes = fread(buf, sizeof(char),
+				      BUFSIZ, form.tmpf)) > 0)
+		    fwrite(buf, sizeof(char), nbytes, fp);
+		} while (nbytes == BUFSIZ);
+	      if (fclose(fp))
+		{
+		  form.errcond=1;
+		  strcpy(form.errmsg, "500 Write to file failed");
+		  cgi_concat_errno(form.errmsg);
+		  strncpy(form.errinfo, incoming_filename, CGI_ERRMSG_MAX);
+		}
+	    }
+	  else
+	    {
+	      form.errcond=1;
+	      strcpy(form.errmsg, "500 Could not append file");
+	      cgi_concat_errno(form.errmsg);
+	      strncpy(form.errinfo, incoming_filename, CGI_ERRMSG_MAX);
+	    }
+	}
+      else
+	{
+	  form.errcond=1;
+	  strcpy(form.errmsg, "500 No write access to file");
+	  cgi_concat_errno(form.errmsg);
+	  strncpy(form.errinfo, incoming_filename, CGI_ERRMSG_MAX);
+	}
+    }
+
+  if (form.errcond)
+    cgi_output_failure(&form, "Form was not processed due to an error.");
+  else
+    cgi_output_success(&form, "The following information was written:");
+  retval = form.errcond;
+  cgi_free_form(&form);
+  return retval;
+}
